/*=============================================================================
    Copyright (c) 2004 Angus Leeming
    Copyright (c) 2007 Joel de Guzman

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef BOOST_PHOENIX_STL_CONTAINER_HPP
#define BOOST_PHOENIX_STL_CONTAINER_HPP

#include <boost/phoenix/stl/container/container.hpp>

#endif // BOOST_PHOENIX_STL_CONTAINER_HPP
